import os
from turbines.config_loader import AppConfig

from abc import ABC, abstractmethod


class PluginBase:
    def __init__(self, config: AppConfig) -> None:
        self.config: AppConfig = config

    def before_build(self):
        pass

    def after_build(self, output_dir: str):
        pass

    def after_page_render(
        self, page_path: str, query_path, metadata: dict, content: str
    ) -> str:

        return content

    def before_page_render(self, page_path: str, content: str) -> str:
        return content


class SitemapGenerator(PluginBase):

    def __init__(self, config: AppConfig) -> None:
        super().__init__(config)
        self._urls = []

    def set_config(self, config):
        self.config = config

    def after_page_render(
        self, page_path: str, query_path: str, metadata: dict, content: str
    ) -> str:

        if metadata.get("noindex", False):
            return content

        site_url = self.config.site.url.rstrip("/")

        page_url = f"{site_url}/{query_path.lstrip('/')}"
        if not self.config.site.sitemap.use_index:
            page_url = page_url.replace("/index.html", "/")
        self._urls.append(page_url)
        return content

    def after_build(self, output_dir: str):
        if not self.config.site.sitemap.enable:
            return

        sitemap_path = os.path.join(output_dir, "sitemap.xml")
        with open(sitemap_path, "w", encoding="utf-8") as f:
            f.write(
                '<?xml version="1.0" encoding="UTF-8"?>\n<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">\n'
            )
            for url in self._urls:
                f.write(f"<url><loc>{url}</loc></url>\n")
            f.write("</urlset>\n")
        print(f"Sitemap generated at {sitemap_path}")

        # Generate Robots.txt with the content specified and the sitemap
        if self.config.site.robots_txt.enable:
            robots_path = os.path.join(output_dir, "robots.txt")
            with open(robots_path, "w", encoding="utf-8") as f:
                if self.config.site.robots_txt.content:
                    f.write(self.config.site.robots_txt.content + "\n")
                f.write(f"Sitemap: {self.config.site.url}/sitemap.xml\n")
            print(f"Robots.txt generated at {robots_path}")



from abc import ABC, abstractmethod

import yaml


class BaseReader(ABC):
    @abstractmethod
    def read(self, filepath) -> tuple[dict, str]:
        with open(filepath, "r", encoding="utf-8") as f:
            return {}, f.read()


class HTMLReader(BaseReader):

    def read(self, filepath) -> tuple[dict, str]:
        with open(filepath, "r", encoding="utf-8") as f:
            metadata = {}
            # if the start of the file is a yaml front matter, parse it
            first_line = f.readline()
            if first_line.strip() == "---":
                # read until the next ---
                front_matter_lines = []
                for line in f:
                    if line.strip() == "---":
                        break
                    front_matter_lines.append(line)
                front_matter = "".join(front_matter_lines)
                metadata = yaml.safe_load(front_matter) or {}
            else:
                f.seek(0)  # reset to start if no front matter
            content = f.read()
            return metadata, content


class MarkdownReader(BaseReader):
    def read(self, filepath) -> tuple[dict, str]:
        import markdown

        with open(filepath, "r", encoding="utf-8") as f:
            md_content = f.read()
        md = markdown.Markdown(extensions=["meta", "extra", "toc"])
        html_content = md.convert(md_content)

        metadata = {}
        for key, value in getattr(md, "Meta", {}).items():
            if isinstance(value, list) and len(value) == 1:
                metadata[key] = value[0]
            else:
                metadata[key] = value

        # Use Jinja2 template inheritance if 'template' is specified in metadata
        if "template" in metadata:
            template_name = metadata["template"]
            # Use Jinja2 block for content and extends for template
            html_content = (
                f"{{% extends '{template_name}' %}}\n"
                "{% block content %}\n"
                f"{html_content}\n"
                "{% endblock %}"
            )

        return metadata, html_content



import typer
from pathlib import Path
from turbines import builder, server

app = typer.Typer()


@app.command()
def create(path: Path):
    """Scaffold a new project structure (pages, templates, static)."""
    builder.scaffold(path)
    print(f"Created project at {path}")


@app.command()
def build():
    """Render pages to the build folder."""

    _builder = builder.Builder()
    _builder.load()
    _builder.build_site()


@app.command()
def serve(
    watch: bool = typer.Option(True, help="Enable hot-reloading"),
    host: str = typer.Option("127.0.0.1", help="Host to bind the server"),
    port: int = typer.Option(8000, help="Port to bind the server"),
):
    """Run local server with hot-reloading."""

    server.run_server(watch=watch, host=host, port=port)


def main() -> None:
    app()



import yaml
from pathlib import Path
from typing import Any
from pydantic import BaseModel, ValidationError


class RobotsTxtConfig(BaseModel):
    enable: bool = False
    content: str | None = None


class SitemapConfig(BaseModel):
    enable: bool = False
    use_index: bool = False


class SiteConfig(BaseModel):
    url: str
    title: str | None = None
    output_dir: str = "dist"
    pages_dir: str = "pages"
    static_dir: str = "static"
    templates_dir: str = "templates"
    robots_txt: RobotsTxtConfig = RobotsTxtConfig()
    sitemap: SitemapConfig = SitemapConfig()


class AppConfig(BaseModel):
    site: SiteConfig
    context: dict[str, Any] = {}


class ConfigLoader:
    @staticmethod
    def load(path: str | Path) -> AppConfig:
        with open(path, "r") as f:
            data: dict[str, Any] = yaml.safe_load(f)
            print(f"Loaded configuration from {path}")
            print(data)

        if data is None:
            data = {}

        try:
            return AppConfig(**data)
        except ValidationError as e:
            raise RuntimeError(f"Invalid configuration: {e}")



import os
import time
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler
from turbines.builder import Builder

import threading
import tornado.ioloop
import tornado.web
import tornado.websocket
import tornado.httpserver


CLIENTS = []
LIVE_RELOAD_SCRIPT = None
MIME_TYPES = {
    ".html": "text/html; charset=UTF-8",
    ".css": "text/css; charset=UTF-8",
    ".js": "application/javascript; charset=UTF-8",
    ".png": "image/png",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".gif": "image/gif",
    ".svg": "image/svg+xml",
}


def make_reload_script(host: str, port: int) -> str:
    return f"""
<script>

    function connectWebSocket() {{
        let ws = new WebSocket("ws://{host}:{port}/_turbines/livereload");
        ws.onmessage = (event) => {{
            if (event.data === "reload") {{
                console.log("Reload message received, reloading page...");
                window.location.reload();
            }} 
            
        }};
        ws.onopen = () => {{
            console.log("LiveReload WebSocket connection established.");
        }};
        ws.onclose = () => {{
            console.log("LiveReload WebSocket connection closed, reconnecting in 1s...");
            setTimeout(connectWebSocket, 5000);
        }};
    }}
    connectWebSocket();
</script>
"""


def notify_client_refresh():
    # print("Notifying clients to reload...")
    for client in list(CLIENTS):
        try:
            client.write_message("reload")
        except:
            CLIENTS.remove(client)


class ChangeHandler(FileSystemEventHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._dist_path = os.path.abspath("dist")
        self._debounce_timer = None
        self._debounce_delay = 0.3  # seconds

    def set_loop(self, loop):
        self.__loop = loop

    def set_builder_ref(self, builder: Builder):
        self._builder = builder
        self._dist_path = os.path.abspath(self._builder.build_path)

    def _handle_change(self, path):

        # Ignore changes in the dist directory
        if path.startswith(self._dist_path):
            return

        print(f"File change detected in {path}, scheduling reload notification...")

        load_static = True if "static" in path else False

        if self.__loop:
            self._builder.reload(
                load_static=load_static,
            )
            self.__loop.add_callback(notify_client_refresh)

    def on_modified(self, event):
        path = os.path.abspath(event.src_path)
        if event.is_directory:
            return
        if self._debounce_timer:
            self._debounce_timer.cancel()
        self._debounce_timer = threading.Timer(
            self._debounce_delay, self._handle_change, args=(path,)
        )
        self._debounce_timer.start()


class LiveReloadWebSocketHandler(tornado.websocket.WebSocketHandler):

    def open(self, *args, **kwargs):
        # print("LiveReload client connected.")
        CLIENTS.append(self)

    def on_close(self):
        CLIENTS.remove(self)

    def on_message(self, message: str | bytes):

        pass

    def check_origin(self, origin: str) -> bool:
        return True  # Allow connections from any origin


class StaticFileHandler(tornado.web.StaticFileHandler):
    async def get(self, path, include_body=True):
        # Serve /index.html for root or empty path
        await super().get(path, include_body)


class StaticFileHandlerWithReload(tornado.web.StaticFileHandler):

    def _inject_reload_script(self, content: str) -> str:
        assert LIVE_RELOAD_SCRIPT is not None, "LIVE_RELOAD_SCRIPT is not set!"
        if "</body>" in content:
            content = content.replace("</body>", LIVE_RELOAD_SCRIPT + "</body>")
        else:
            content += LIVE_RELOAD_SCRIPT
        return content

    async def get(self, path, include_body=True):
        if path == "" or path.endswith("/"):
            path = os.path.join(path, str("index.html"))
        elif path.endswith(".html"):
            pass
        else:
            await super().get(path, include_body)
            return

        with open(path, "r") as f:
            content = f.read()
            self.set_header("Content-Type", "text/html; charset=UTF-8")
            content = self._inject_reload_script(content)
            self.write(content)
            await self.flush()


class TurbineServer:
    def __init__(self, watch: bool = False):
        self.watch = watch
        self.builder = Builder(inject_reload_script=True)
        self.builder.load()
        self.builder.build_site()

    def serve(self, host: str, port: int):
        os.chdir(self.builder.build_path)
        print(f"Serving '{self.builder.build_path}' at http://{host}:{port} ...")
        print("Do not use in production!")

        # set up live reload script
        global LIVE_RELOAD_SCRIPT
        LIVE_RELOAD_SCRIPT = make_reload_script(host, port)

        handler = StaticFileHandlerWithReload
        if not self.watch:
            handler = StaticFileHandler

        self.app = tornado.web.Application(
            [
                (r"/_turbines/livereload", LiveReloadWebSocketHandler),
                (
                    r"/(.*)",
                    handler,
                    {"path": self.builder.build_path, "default_filename": "index.html"},
                ),
            ]
        )
        server = tornado.httpserver.HTTPServer(self.app)
        server.listen(port, address=host)

    def run(self, host: str = "localhost", port: int = 8000):
        loop = tornado.ioloop.IOLoop.current()
        observer = None
        if self.watch:
            observer = Observer()
            handler = ChangeHandler()
            handler.set_loop(loop)
            handler.set_builder_ref(self.builder)
            observer.schedule(handler, path=os.path.join(os.getcwd()), recursive=True)
            observer.start()
        try:
            self.serve(host, port)
            tornado.ioloop.IOLoop.current().start()
        finally:
            if observer:
                observer.stop()
                observer.join()


def run_server(watch: bool = False, host: str = "localhost", port: int = 8000):
    server = TurbineServer(watch=watch)
    server.run(host, port)



import os
import shutil
from datetime import datetime
from typing import Type
from jinja2 import Environment, FileSystemLoader, select_autoescape
from jinja2_simple_tags import StandaloneTag

from turbines.config_loader import AppConfig, ConfigLoader
from turbines.index_tools import SitemapGenerator
from turbines.reader import BaseReader, HTMLReader, MarkdownReader


class NowExtension(StandaloneTag):
    tags = {"now"}

    def render(self, format="%Y-%m-%d %H:%I:%S"):
        return datetime.now().strftime(format)


class StaticFileExtension(StandaloneTag):
    tags = {"static"}

    def render(self, filename):
        return f"/static/{filename}"


def scaffold(path):
    # make a diretory in the specified path if it doesn't exist
    if not os.path.exists(path):
        os.makedirs(path)
        print(f"Created directory at {path}")
    else:
        print(f"Directory already exists at {path}")

    # copy ./scaffold to the specified path
    scaffold_src = os.path.join(os.path.dirname(__file__), "scaffold")
    scaffold_dst = os.path.join(path)

    # copy the data from scaffold_src to scaffold_dst

    shutil.copytree(scaffold_src, scaffold_dst, dirs_exist_ok=True)
    print(f"Copied scaffold to {path}")


class Builder:

    def __init__(self, inject_reload_script: bool = False):
        self.config: AppConfig | None = None
        self.static_files: dict[str, str] = {}
        self.inject_reload_script = inject_reload_script

    def load(self):
        self.config = self.load_config()

        self.build_path = os.path.join(os.getcwd(), self.config.site.output_dir)
        os.makedirs(self.build_path, exist_ok=True)

        self.static_path = os.path.join(os.getcwd(), self.config.site.static_dir)

        self.load_static(self.static_path)
        self.load_templates(self.config.site.templates_dir)
        self.load_pages(self.config.site.pages_dir)

        self.pages_path = os.path.join(os.getcwd(), self.config.site.pages_dir)
        self.templates_path = os.path.join(os.getcwd(), self.config.site.templates_dir)

        self.global_context = self.config.context or {}

        # TODO temporary sitemap plugin setup
        sitemap_plugin = SitemapGenerator(self.config)

        self.plugins = [sitemap_plugin]

    def load_config(self):
        self.config_path = os.path.join(os.getcwd(), "config.yaml")

        if os.path.isfile(self.config_path):
            print("Found config.yml")
        else:
            print("config.yml not found")

        try:
            config = ConfigLoader.load(self.config_path)
        except RuntimeError as e:
            print(f"Error loading config: {e}")
            raise e

        return config

    def load_pages(self, pages_path):

        for root, _, files in os.walk(pages_path):
            for filename in files:
                file_path = os.path.join(root, filename)
                # For now, just print the page paths
                # print(f"Found page: {os.path.relpath(file_path, pages_path)}")

    def load_static(self, static_path):
        # Copy static files to <build_path>/static
        output_static_path = os.path.join(self.build_path, "static")
        if os.path.isdir(static_path):
            shutil.copytree(static_path, output_static_path, dirs_exist_ok=True)

    def load_templates(self, templates_path):
        pass

    def reload(self, load_static: bool = False):
        if load_static:
            self.load_static(self.static_path)
        self.build_site()

    def build_site(self):

        # Run plugin before build hook
        for plugin in self.plugins:
            plugin.before_build()

        if self.config is None:
            raise RuntimeError("Config not loaded. Call load() before build_site().")

        # Set up Jinja2 environment
        env = Environment(
            loader=FileSystemLoader([self.pages_path, self.templates_path]),
            autoescape=select_autoescape(["html", "xml"]),
        )

        env.globals["context"] = self.global_context
        env.globals["site"] = {
            "title": self.config.site.title,
            "url": self.config.site.url,
        }

        # add the now tag
        env.add_extension(NowExtension)
        env.add_extension(StaticFileExtension)

        READERS: dict[str, Type[BaseReader]] = {
            ".html": HTMLReader,
            ".htm": HTMLReader,
            ".md": MarkdownReader,
        }

        # Render each page in ./pages
        if not os.path.isdir(self.pages_path):
            print("No pages to render.")
            return

        for root, _, files in os.walk(self.pages_path):
            rel_root = os.path.relpath(root, self.pages_path)

            for filename in files:
                file_ext = os.path.splitext(filename)[-1].lower()
                ReaderClass = READERS.get(file_ext)

                if not ReaderClass:
                    print(f"Skipping unsupported file type: {filename}")
                    continue

                reader = ReaderClass()
                file_path = os.path.join(root, filename)
                metadata, content = reader.read(file_path)

                # create the rendered output using jinja from the content
                template = env.from_string(content)
                rendered = template.render(**metadata)

                name_without_ext = os.path.splitext(filename)[0]
                # Preserve directory structure in output
                output_dir = os.path.join(self.build_path, rel_root)
                os.makedirs(output_dir, exist_ok=True)
                output_path = os.path.join(output_dir, name_without_ext + ".html")

                # Remove the build directory from output_path to get the query_path
                query_path = os.path.relpath(output_path, self.build_path)

                for plugin in self.plugins:
                    rendered = plugin.after_page_render(
                        file_path, query_path, metadata, rendered
                    )

                with open(output_path, "w", encoding="utf-8") as out_f:
                    out_f.write(rendered)
                print(
                    f"    Rendered {os.path.relpath(file_path, self.pages_path)} -> {query_path}"
                )

        # Run plugin after build hook
        for plugin in self.plugins:
            plugin.after_build(self.build_path)
